/*!
 *  Epiclock Styles
 *
 *  Copyright (c) Eric Garside
 *  Dual licensed under:
 *      MIT: http://www.opensource.org/licenses/mit-license.php
 *      GPLv3: http://www.opensource.org/licenses/gpl-3.0.html
 */
 
/*  Generic Styles
----------------------------------*/

.epiclock-wait-for-render
    { visibility: hidden; }

.epiclock-spacer
    { padding-right: 3px; }