# Metadata for the widget EGG. Should leave it for the (unimplemented) widget
# tracker to track it

__DISTRIBUTION__ = 'tw.epiclock'
__DESCRIPTION__ = 'Insert Clocks in templates'
__URL__ = 'http://code.google.com/p/twepiclock'
__VERSION__ = '0.1a1'
__AUTHOR__ = 'Nicolas Laurance'
__EMAIL__ = 'nicolas[dot]laurance<at>gmail'
__COPYRIGHT__ = "Copyright 2010 Nicolas Laurance"
__LICENSE__ = 'MIT GPLv3'
